import time
import nibabel as nib
import numpy as np

from lyroi.utils import get_model_folders, get_folds, get_tmp_dir, get_suffixes, validate_extensions
from lyroi.nnunet_interface import nnunet_predict, get_torch_device
from pathlib import Path
from shutil import move

def merge_delineations(input_folders, output_folder, strategy="u", force = True):
    assert strategy in ["u", "i", "m"], "Invalid merging strategy"

    input_files = [list(Path(input_dir).glob("*.nii.gz")) for input_dir in input_folders]
    assert len(set([len(l) for l in input_files])) == 1, "Number of images in the submodel output folders do not match. Something went wrong with the predictions"

    input_basenames = [[input_file.name for input_file in input_folder] for input_folder in input_files]
    __ = [input_dir.sort() for input_dir in input_basenames]  # sort lists
    assert input_basenames.count(input_basenames[0]) == len(input_basenames), "Files in the submodel output folders do not match. Something went wrong with the predictions"

    for file_name in input_basenames[0]:
        files_in = [Path(input_dir, file_name) for input_dir in input_folders]
        file_out = Path(output_folder, file_name)

        if file_out.exists():
            if force:
                file_out.unlink()
            else:
                raise FileExistsError(f"Output file {file_out} already exists!")

        if len(files_in) == 1:
            # no need to merge anything. Just move files
            move(files_in[0], file_out)
        else:
            # okay, now we actually need to read files and save results
            imgs_in = [nib.load(file) for file in files_in]
            vols_in = [nifti.get_fdata() for nifti in imgs_in]

            if strategy == "u":
                result = np.logical_or.reduce(vols_in)
            if strategy == "i":
                result = np.logical_and.reduce(vols_in)
            if strategy == "m":
                result = np.average(vols_in, axis=0) > 0.5
            nifti_out = nib.Nifti1Image(result.astype(np.uint8), affine=imgs_in[0].affine, header=imgs_in[0].header)
            nib.save(nifti_out, file_out)

def check_inputs(input_folder, mode):
    suffixes = get_suffixes(mode)

    if len(suffixes) == 1:
        return True

    suffixes = [suffix + ".nii.gz" for suffix in suffixes]
    input_lists = [list(Path(input_folder).glob("*"+ suffix)) for suffix in suffixes]
    input_ids = [[input_file.name.removesuffix(suffix) for input_file in input_list] for input_list, suffix in zip(input_lists, suffixes)]

    unpaired = (set(input_ids[0]) - set(input_ids[1])) | (set(input_ids[1]) - set(input_ids[0]))
    if len(unpaired) == 0:
        return True
    else:
        names = "\n".join(unpaired)
        exit("Cannot proceed, the following patients are missing either CT or PET:\n" + names)

def transfer_input_files(input_files, target_folder, mode, pname = 'patient_001'):
    suffixes = get_suffixes(mode)
    n_channels = len(suffixes)
    if len(input_files) != n_channels:
        exit(f"Number of input files does not match the number of input channels for the selected mode ({n_channels})")
    for input_file, suffix in zip(input_files, suffixes):
        Path(target_folder, pname + suffix + ".nii.gz").symlink_to(input_file)

def transfer_output_files(input_folder, output_file, pname = 'patient_001'):
    Path(output_file).unlink(missing_ok=True)
    move(Path(input_folder, pname + ".nii.gz"), output_file)

def predict_from_folder(input_folder, output_folder, mode, device='gpu'):
    check_inputs(input_folder, mode)

    model_folders = get_model_folders(mode)
    folds = get_folds(mode)
    tmp_dir = get_tmp_dir()
    tmp_rundir = Path(tmp_dir, str(time.time_ns()))
    tmp_subdirs = []

    print("Starting predictions. Wait until all models finish prediction to see the results")
    torch_device = get_torch_device(device)
    start_time = time.time()
    try:
        counter = 0
        for folder in model_folders:
            counter += 1
            print(f"Predicting with model {counter}/{len(model_folders)}")
            tmp_subdir = Path(tmp_rundir, Path(folder).stem)
            tmp_subdirs.append(tmp_subdir)
            nnunet_predict(input_folder, tmp_subdir, folder, folds, torch_device)

        # import multiprocessing as mp
        # mp.set_start_method("spawn", force=True)
        #
        # procs = []
        # for i, folder in enumerate(model_folders):
        #     if i > 0:
        #         time.sleep(20)
        #     tmp_subdir = Path(tmp_rundir, Path(folder).stem)
        #     tmp_subdirs.append(tmp_subdir)
        #     p = mp.Process(
        #         target=nnunet_predict,
        #         args=(input_folder, tmp_subdir, folder, folds, torch_device)
        #     )
        #     p.start()
        #     procs.append(p)
        #
        #
        # for p in procs:
        #     p.join()
        print("Merging delineations...")
        merge_delineations(tmp_subdirs, output_folder)
        print("Execution time: %s seconds " % (time.time() - start_time))
    except Exception as e:
        print("Execution halted: ", e.args[0])
        raise e
    finally:
        for tmp_subdir in tmp_subdirs:
            for file in tmp_subdir.iterdir():
                file.unlink()
            tmp_subdir.rmdir()
        tmp_rundir.rmdir()
        try:
            tmp_dir.rmdir()
        except Exception:
            pass

def predict_from_files(input_files, output_file, mode, device='gpu'):
    validate_extensions(input_files + [output_file], ".nii.gz")

    tmp_dir = get_tmp_dir()
    tmp_rundir = Path(tmp_dir, str(time.time_ns()))
    tmp_input_dir = Path(tmp_rundir, "input")
    tmp_output_dir = Path(tmp_rundir, "output")
    try:
        tmp_input_dir.mkdir(exist_ok=True, parents=True)
        tmp_output_dir.mkdir(exist_ok=True, parents=True)
        transfer_input_files(input_files, tmp_input_dir, mode)
        predict_from_folder(tmp_input_dir, tmp_output_dir, mode, device)
        transfer_output_files(tmp_output_dir, output_file)
    except Exception as e:
        print("Execution halted: ", e.args[0])
        raise e
    finally:
        for file in tmp_input_dir.iterdir():
            file.unlink()
        for file in tmp_output_dir.iterdir():
            file.unlink()
        tmp_input_dir.rmdir()
        tmp_output_dir.rmdir()
        tmp_rundir.rmdir()
        try:
            tmp_dir.rmdir()
        except Exception:
            pass



# Parts taken and modified after https://github.com/MIC-DKFZ/nnUNet/
import os

import psutil
import torch
from nnunetv2.inference.predict_from_raw_data import nnUNetPredictor


def get_torch_device(device='gpu'):
    assert device in ['cpu', 'cpu-max', 'gpu',
                      'mps'], f'-device must be either cpu, cpu-max, gpu or mps. Other devices are not tested/supported. Got: {device}'
    if device == 'cpu':
        torch.set_num_threads(8)
        device = torch.device('cpu')
    if device == 'cpu-max':
        torch.set_num_threads(psutil.cpu_count(logical=False))
        device = torch.device('cpu')
    if device == 'gpu':
        torch.set_num_threads(1)
        device = torch.device('cuda')
    if device == 'mps':
        device = torch.device('mps')

    return device

def nnunet_predict(input_folder, output_folder, model_folder, folds, torch_device):
    predictor = nnUNetPredictor(tile_step_size=0.5,
                                use_gaussian=True,
                                use_mirroring=True,
                                perform_everything_on_device=True,
                                device=torch_device,
                                verbose=False,
                                verbose_preprocessing=False,
                                allow_tqdm=True)

    predictor.initialize_from_trained_model_folder(
        model_folder,
        folds,
        checkpoint_name='checkpoint_final.pth'
    )

    predictor.predict_from_files(str(input_folder), str(output_folder),
                                 save_probabilities=False,
                                 overwrite=True,
                                 num_processes_preprocessing=3,
                                 num_processes_segmentation_export=3,
                                 folder_with_segs_from_prev_stage=None,
                                 num_parts=1,
                                 part_id=0)


import json
import os
import sys
import math
import requests
import psutil

from pathlib import Path


def validate_extensions(file_list, expected_ext = ".nii.gz"):
    return all([file.endswith(expected_ext) for file in file_list])

def yes_no_input(prompt, no_message):
    answer = input(prompt + " ([y]/n)? ")
    if answer.lower() in ['y', 'yes', '']:
        return
    elif answer.lower() in ['n', 'no']:
        print(no_message)
        sys.exit(0)
    else:
        print("Incorrect input, try again")
        yes_no_input(prompt, no_message)

# taken from https://stackoverflow.com/questions/5194057/better-way-to-convert-file-sizes-in-python
def format_file_size(size_bytes):
   if size_bytes == 0:
       return "0B"
   size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
   i = int(math.floor(math.log(size_bytes, 1024)))
   p = math.pow(1024, i)
   s = round(size_bytes / p, 2)
   return "%s %s" % (s, size_name[i])

def get_model_folders(mode):
    from nnunetv2.utilities.file_path_utilities import get_output_folder
    folder_list = []
    if mode == 'petct':
        folder_list.append(
            get_output_folder(1, 'nnUNetTrainer', 'nnUNetPlans', '3d_fullres')
        )
        folder_list.append(
            get_output_folder(1, 'nnUNetTrainer', 'nnUNetResEncUNetMPlans', '3d_fullres')
        )
        folder_list.append(
            get_output_folder(1, 'nnUNetTrainer', 'nnUNetResEncUNetLPlans', '3d_fullres')
        )
    return folder_list

def get_folds(mode):
    folds = []
    if mode == 'petct':
        folds = [0, 1, 2, 3, 4]
    if mode == 'petct_turbo':
        folds = ['all']
    return folds

def get_suffixes(mode):
    suffixes = []
    if mode == 'petct':
        suffixes = ['_0000', '_0001']
    if mode == 'petct_turbo':
        suffixes = ['_0000', '_0001']
    return suffixes

def get_repository_url():
    try:
        # collection of sources. Should automatically resolve to the latest versions of the models
        r = requests.head("https://rodare.hzdr.de/record/4160", allow_redirects=True)
        return r.url
    except Exception as e:
        exit("Cannot reach the online model repository! Please check your internet connection or contact the developer.")

def get_download_urls(mode, repository_url = None):
    if repository_url is None:
        repository_url = get_repository_url()

    download_urls = []
    if mode == 'petct':
        download_urls.append(repository_url + "/files/LyROI_Orig.zip")
        download_urls.append(repository_url + "/files/LyROI_ResM.zip")
        download_urls.append(repository_url + "/files/LyROI_ResL.zip")
    return download_urls

def get_download_size(mode):
    download_urls = get_download_urls(mode)
    try:
        size = 0
        for url in download_urls:
            r = requests.head(url)
            size += int(r.headers['content-length'])
        return size
    except Exception as e:
        exit("Cannot locate the model files! Please check your internet connection or contact the developer.")

def install_model(mode):
    from nnunetv2.model_sharing.model_download import download_and_install_from_url
    repository_url = get_repository_url()
    ver = check_version_online(mode, repository_url)

    # downloading and installing
    download_urls = get_download_urls(mode, repository_url)
    print("Wait until all downloads are finished")
    for url in download_urls:
        download_and_install_from_url(url)

    # writing a current version
    model_folders = get_model_folders(mode)
    for folder in model_folders:
        Path(folder, "VERSION").write_text(ver)

def check_version_online(mode, repository_url = None):
    if repository_url is None:
        repository_url = get_repository_url()
    try:
        r = requests.get(repository_url + "/files/VERSION")
        version_file = r.content.decode("utf-8")
        j_file = json.loads(version_file)
        return j_file[mode]
    except Exception as e:
        exit("Cannot find version info in the online model repository! Please contact the developer.")

def check_version_local(mode):
    version_list = []
    model_folders = get_model_folders(mode)
    for folder in model_folders:
        if Path(folder, "VERSION").exists():
            version_list.append(Path(folder, "VERSION").read_text())
        else:
            version_list.append(None)
    uniques = set(version_list)

    if None in uniques:
        raise Exception("One or more submodels are missing the version information!")
    if len(uniques) > 1:
        raise Exception("Versions mismatch between the submodels!")

    return uniques.pop()

def check_model(mode):
    try:
        folder_list = get_model_folders(mode)
    except Exception as e:
        return False
    fold_list = get_folds(mode)
    status = True
    for folder in folder_list:
        status = status and Path(folder, "dataset.json").exists()
        status = status and Path(folder, "plans.json").exists()
        for fold in fold_list:
            status = status and Path(folder, "fold_" + str(fold), "checkpoint_final.pth").exists()
    return status

def get_lyroi_dir():
    if 'LYROI_DIR' in os.environ:
        return os.environ['LYROI_DIR']
    else:
        return str(Path(Path.home() , '.lyroi')) #default

def get_models_dir():
    return str(Path(get_lyroi_dir(), "nnUNet_results"))

def get_tmp_dir():
    return str(Path(get_lyroi_dir(), "tmp"))

def check_setup():
    status = True
    lyroi_dir = get_lyroi_dir()
    models_dir = get_models_dir()
    status = status and Path(lyroi_dir).exists()
    status = status and 'nnUNet_raw' in os.environ
    status = status and 'nnUNet_preprocessed' in os.environ
    status = status and 'nnUNet_results' in os.environ
    status = status and Path(lyroi_dir) in Path(models_dir).parents
    return status

def setup_lyroi():
    lyroi_dir = get_lyroi_dir()
    if not Path(lyroi_dir).exists():
        print("Creating LyROI directory: " + lyroi_dir)
        Path(lyroi_dir).mkdir(exist_ok=True, parents=True)

    models_dir = get_models_dir()
    if not Path(models_dir).exists():
        Path(models_dir).mkdir(exist_ok=True, parents=True)

    os.environ['nnUNet_raw'] = ""
    os.environ['nnUNet_preprocessed'] = ""
    os.environ['nnUNet_results'] = models_dir # only this one matters
    # some performance tweeks:
    os.environ['nnUNet_def_n_proc'] =(
        str(psutil.cpu_count(logical=False))) if 'nnUNet_def_n_proc' not in os.environ \
        else int(os.environ['nnUNet_def_n_proc'])
    os.environ['OMP_NUM_THREADS'] = str(psutil.cpu_count(logical=False))
    os.environ['MKL_NUM_THREADS'] = str(psutil.cpu_count(logical=False))
    os.environ['OMP_PROC_BIND'] = "close"
    os.environ['OMP_PLACES'] = "cores"
    os.environ['OMP_SCHEDULE'] = "static"
    os.environ['GOMP_CPU_AFFINITY '] = "N-M"


if __name__ == "__main__":
    setup_lyroi()
    print(os.environ['nnUNet_results'])
    print("All correct:", check_setup())
    print("Model found:", check_model("petct"))
    print("Online version:", check_version_online("petct"))
    try:
        version = check_version_local("petct")
        print("Local version:", version)
    except Exception as e:
        print(e)